import base64
import io
import itertools
import os
import shutil
import logging
from datetime import datetime, timedelta, date
from decimal import Decimal
from uuid import UUID
from zipfile import ZipFile

from PySimpleGUI import POPUP_BUTTONS_OK_CANCEL, PySimpleGUI
from markdown2 import markdown
from satcfdi import DatePeriod, CFDI, Code, Signer
from satcfdi.accounting import filter_invoices_iter, SatCFDI
from satcfdi.create import Issuer
from satcfdi.exceptions import ResponseError, DocumentNotFoundError
from satcfdi.pacs import Accept
from satcfdi.pacs.sat import SAT, TipoDescargaMasivaTerceros, EstadoSolicitud
# noinspection PyUnresolvedReferences
from satcfdi.transform.catalog import CATALOGS

from weasyprint import HTML, CSS

from . import SOURCE_DIRECTORY, __version__, local
from .client_validation import validar_client
from .file_data_managers import environment_default, environment_bold_escaped, ClientsManager, ConfigManager
from .formatting_functions.common import fecha, pesos, porcentaje
from .gui_functions import generate_ingresos, pago_factura, find_ajustes, format_concepto_desc, exportar_facturas, exportar_facturas_filename, parse_ym_date
from .layout import make_layout, InvoiceButtonManager, EmailButtonManager
from .log_tools import LogHandler, LogAdapter, log_line, log_item, cfdi_header
from .mycfdi import get_all_cfdi, MyCFDI, move_to_folder, local_db, PPD, PUE

logging.basicConfig(level=logging.INFO)
logging.getLogger("weasyprint").setLevel(logging.ERROR)
logging.getLogger("fontTools").setLevel(logging.ERROR)

logger = LogAdapter(logging.getLogger())

AJUSTES_DIR = "ajustes"
template = environment_default.get_template(
    'mail_facturas_template.html'
)


class FacturacionGUI:
    def __init__(self, pac_service, csd_signer: Signer, email_manager, fiel_signer=None, concepto_adapter=None):
        self.email_manager = email_manager
        self.pac_service = pac_service
        self.sat_service = SAT(
            signer=fiel_signer
        )
        self.concepto_adapter = concepto_adapter
        clients = ClientsManager()

        self.issuer_cif = clients[csd_signer.rfc]
        self.lugar_expedicion = self.issuer_cif['CodigoPostal']
        self.issuer = Issuer(
            signer=csd_signer,
            tax_system=self.issuer_cif['RegimenFiscal'],
            legal_name=self.issuer_cif['RazonSocial']
        )

        MyCFDI.issuer_rfc = self.issuer.rfc
        window = PySimpleGUI.Window(
            f"Facturacion 4.0  RazonSocial: {self.issuer.legal_name}  RFC: {self.issuer.rfc}  "
            f"Serie: {local.config.serie()}  Regimen: {self.issuer.tax_system}  LugarExpedicion: {self.lugar_expedicion}",
            make_layout(fiel_signer is not None),
            size=(1280, 800),
            resizable=True,
        )
        self.window = window

        self.ch = LogHandler(self.window['console'])
        self.ch.setLevel(logging.INFO)
        logging.getLogger().addHandler(self.ch)

        self.invoice_button_manager = InvoiceButtonManager(window["crear_facturas"], window["detallado"])
        self.email_button_manager = EmailButtonManager(window["enviar_correos"])
        self.all_invoices = None

    def initial_screen(self):
        def convert_ans1_date(ans1_date):
            return datetime.strptime(ans1_date.decode('utf-8'), '%Y%m%d%H%M%SZ')

        def cert_info(signer: Signer):
            if signer:
                return {
                    "NoCertificado": signer.certificate_number,
                    "Tipo": str(signer.type),

                    "organizationName": signer.certificate.get_subject().O,
                    "x500UniqueIdentifier": signer.certificate.get_subject().x500UniqueIdentifier,
                    "serialNumber": signer.certificate.get_subject().serialNumber,
                    "organizationUnitName": signer.certificate.get_subject().OU,
                    "emailAddress": signer.certificate.get_subject().emailAddress,

                    "notAfter": convert_ans1_date(signer.certificate.get_notAfter()),
                    "notBefore": convert_ans1_date(signer.certificate.get_notBefore()),
                }

        try:
            logger.info_yaml({
                "version": __version__.__version__,
                "facturacion": "CFDI 4.0",
                "emisor": self.issuer_cif,
                "pac_service": {
                    "Type": type(self.pac_service).__name__,
                    "Rfc": self.pac_service.RFC,
                    "Environment": str(self.pac_service.environment)
                },
                "csd_signer": cert_info(self.issuer.signer),
                "fiel_signer": cert_info(self.sat_service.signer),
                "config": local.config,
            })
        except Exception as ex:
            logger.exception("Error al obtener datos")

    def start(self):
        self.window.finalize()
        self.initial_screen()
        self.main_loop()
        self.window.close()

    def get_all_invoices(self):
        if self.all_invoices:
            return self.all_invoices
        self.all_invoices = get_all_cfdi()
        return self.all_invoices

    def generate_invoice(self, invoice, ref_id=None):
        res = self.pac_service.stamp(
            cfdi=invoice,
            accept=Accept.XML_PDF,
            ref_id=ref_id
        )
        local.config.inc_folio()
        return move_to_folder(res.xml, pdf_data=res.pdf)

    def enviar_correos(self, emails):
        with self.email_manager.sender as s:
            for receptor, notify_invoices, facturas_pendientes in emails:
                receptor_nombre = receptor["RazonSocial"]
                to_addrs = receptor["Email"]

                attachments = []
                for r in notify_invoices:
                    attachments += [r.filename + ".xml", r.filename + ".pdf"]

                message = self.generate_html_template(
                    'mail_facturas_template.html',
                    fields={
                        'invoices': notify_invoices,
                        'pending_invoices': facturas_pendientes,
                    }
                )

                s.send_email(
                    subject=f"Comprobantes Fiscales {receptor_nombre} - {receptor['Rfc']}",
                    to_addrs=to_addrs,
                    html=message,
                    file_attachments=attachments
                )

                self.window.read(timeout=0)

                for r in notify_invoices:
                    r.notified = ",".join(to_addrs)

    def recupera_comprobantes(self, id_solicitud):
        response = self.sat_service.recover_comprobante_status(
            id_solicitud=id_solicitud
        )
        logger.info_yaml(response)
        self.window.read(timeout=0)
        if response["EstadoSolicitud"] == EstadoSolicitud.Terminada:
            for id_paquete in response['IdsPaquetes']:
                response, paquete = self.sat_service.recover_comprobante_download(
                    id_paquete=id_paquete
                )
                logger.info_yaml(response)
                self.window.read(timeout=0)
                yield id_paquete, base64.b64decode(paquete) if paquete else None

    def unzip_cfdi(self, file):
        with ZipFile(file, "r") as zf:
            for fileinfo in zf.infolist():
                xml_data = zf.read(fileinfo)
                move_to_folder(xml_data, pdf_data=None)
                self.window.read(timeout=0)

    def generate_pdf_template(self, template_name, fields):
        increment_template = environment_bold_escaped.get_template(template_name)
        md5_document = increment_template.render(
            emisor=self.issuer_cif,
            fecha_hoy=fecha(date.today()),
            **fields
        )
        html = markdown(md5_document)
        pdf = HTML(string=html).write_pdf(
            target=None,
            stylesheets=[
                os.path.join(SOURCE_DIRECTORY, "markdown_styles", "markdown6.css"),
                CSS(
                    string='@page { width: Letter; margin: 1.6cm 1.6cm 1.6cm 1.6cm; }'
                )
            ]
        )
        return pdf

    def generate_html_template(self, template_name, fields):
        increment_template = environment_default.get_template(template_name)
        render = increment_template.render(
            emisor=self.issuer_cif, **fields
        )
        return render

    def factura_uuid(self, text):
        try:
            factura_uuid = UUID(text)
            factura = self.get_all_invoices().get(factura_uuid)
            if factura:
                return factura
        except:
            pass

    def factura_buscar(self, text):
        if text:
            emisor_rfc = self.issuer.rfc
            res = None
            for i in self.get_all_invoices().values():
                if i.name == text and i["Emisor"]["Rfc"] == emisor_rfc:
                    logger.info(f"Factura Encontrada: {i['Receptor']['Rfc']}  {i.name}  {i.uuid}  {i['Fecha']}")
                    if res:
                        logger.info(f"Multiples Facturas Encontradas con el mismo nombre: {text}")
                        return
                    res = i
            if res:
                return res
        logger.info(f"Factura No Encontrada {text}")

    def log_cfdi(self, cfdi: CFDI):
        cfdi_copy = cfdi.copy()
        del cfdi_copy["Certificado"]
        del cfdi_copy["Sello"]

        if not self.window['detallado'].get():
            del cfdi_copy["Serie"]
            del cfdi_copy["NoCertificado"]
            cfdi_copy.pop("Emisor")
            cfdi_copy["Receptor"] = f"{cfdi_copy['Receptor']['Rfc']}, {cfdi_copy['Receptor']['Nombre']}, {cfdi_copy['Receptor'].get('RegimenFiscalReceptor')}"
            cfdi_copy["Conceptos"] = [x['Descripcion'] for x in cfdi_copy["Conceptos"]]  # f"<< {len(cfdi_copy['Conceptos'])} >>"
            cfdi_copy.pop("Impuestos", None)
            cfdi_copy.pop("Fecha")
            cfdi_copy.pop("LugarExpedicion")
            cfdi_copy.pop("Version")
            cfdi_copy.pop("TipoDeComprobante")
            if cfdi_copy.get("Exportacion") == "01":
                del cfdi_copy["Exportacion"]
            if cfdi_copy.get("FormaPago") == "99":
                del cfdi_copy["FormaPago"]
            if cfdi_copy.get("Moneda") in ("MXN", "XXX"):
                del cfdi_copy["Moneda"]

            if complemento := cfdi_copy.get("Complemento"):
                cfdi_copy["Complemento"] = {}
                if pagos := complemento.get("Pagos"):
                    def clean_doc(d):
                        d = d.copy()
                        d.pop("ImpuestosDR")
                        if d.get("MonedaDR") in ("MXN", "XXX"):
                            del d["MonedaDR"]
                            d.pop("EquivalenciaDR")
                        return d

                    def clean_pago(p):
                        p = p.copy()
                        p.pop("ImpuestosP")
                        p["DoctoRelacionado"] = [clean_doc(x) for x in p["DoctoRelacionado"]]
                        if p.get("MonedaP") in ("MXN", "XXX"):
                            del p["MonedaP"]
                            p.pop("TipoCambioP")
                        return p

                    pagos_copy = pagos.copy()
                    pagos_copy["Pago"] = [clean_pago(x) for x in pagos_copy["Pago"]]
                    cfdi_copy["Complemento"]["Pagos"] = pagos_copy

                if timbre_fiscal_digital := complemento.get("TimbreFiscalDigital"):
                    cfdi_copy["Complemento"]["TimbreFiscalDigital"] = timbre_fiscal_digital['UUID']

        if isinstance(cfdi, SatCFDI):
            cfdi_copy["_saldo_pendiente"] = cfdi.saldo_pendiente

        logger.info_yaml(cfdi_copy)

    def main_loop(self):
        factura_seleccionada = None  # type: MyCFDI | None

        while True:
            event, values = self.window.read()
            local.config = ConfigManager()
            try:
                if event in ("Exit", PySimpleGUI.WIN_CLOSED):
                    return

                if event not in ("crear_facturas", "enviar_correos", "confirm_pago_button", "ver_factura", "ver_excel"):
                    self.ch.clear()
                invoices_to_create = self.invoice_button_manager.clear()
                emails_to_send = self.email_button_manager.clear()

                match event:
                    case "about":
                        self.initial_screen()

                    case "factura_pagar" | "buscar_factura":
                        text = self.window["factura_pagar"].get().strip()
                        if event == "buscar_factura":
                            factura_seleccionada = self.factura_uuid(text)
                            if not factura_seleccionada:
                                factura_seleccionada = self.factura_buscar(text)
                        else:
                            factura_seleccionada = self.factura_uuid(text)

                        if factura_seleccionada:
                            self.log_cfdi(factura_seleccionada)
                            local_db.describe(factura_seleccionada)

                        not_ppd_all = not factura_seleccionada or factura_seleccionada.get("MetodoPago") != PPD or factura_seleccionada.estatus != "1"
                        not_ppd = not_ppd_all or factura_seleccionada["Emisor"]["Rfc"] != self.issuer.rfc

                        not_pue = not factura_seleccionada or factura_seleccionada.get("MetodoPago") != PUE or factura_seleccionada.estatus != "1" or \
                                  factura_seleccionada["Emisor"]["Rfc"] != self.issuer.rfc

                        self.window["status_sat"].update(disabled=not factura_seleccionada)
                        self.window["pago_pue"].update(disabled=not_pue)
                        self.window["email_notificada"].update(
                            disabled=not factura_seleccionada
                                     or factura_seleccionada["Emisor"]["Rfc"] != self.issuer.rfc
                        )
                        self.window["ver_factura"].update(disabled=not factura_seleccionada)

                        self.window["ignorar_ppd"].update(disabled=not_ppd_all)
                        self.window["prepare_pago"].update(disabled=not_ppd)
                        self.window["fecha_pago_select"].update(disabled=not_ppd)
                        self.window["fecha_pago"].update(disabled=not_ppd)
                        self.window["forma_pago"].update(disabled=not_ppd)

                        def is_uuid(s):
                            try:
                                UUID(s)
                                return True
                            except ValueError:
                                return False

                        self.window["descarga"].update(disabled=not is_uuid(text))

                    case "preparar_ajuste_anual":
                        log_item(f"AJUSTE ANUAL")
                        ym_date = parse_ym_date(values['anio_mes_ajuste'])
                        ajuste_porcentaje = values['ajuste_porcentaje']
                        if not ajuste_porcentaje:
                            logger.error("Especificar Ajuste Porcetaje")
                            continue

                        ajuste_porcentaje = Decimal(ajuste_porcentaje) / 100

                        shutil.rmtree(AJUSTES_DIR, ignore_errors=True)
                        os.makedirs(AJUSTES_DIR, exist_ok=True)
                        clients = ClientsManager()

                        has_ajustes = False
                        for i, (rfc, concepto) in enumerate(find_ajustes(ym_date.month), start=1):
                            has_ajustes = True
                            receptor = clients[rfc]
                            valor_unitario_nuevo = concepto["ValorUnitario"] * (1 + ajuste_porcentaje)
                            format_concepto_desc(concepto, periodo="INMUEBLE")

                            data = {
                                "receptor": receptor,
                                "concepto": concepto,
                                "valor_unitario_nuevo": pesos(valor_unitario_nuevo),
                                "ajuste_porcentaje": porcentaje(ajuste_porcentaje, 2),
                                "ajuste_periodo": "UN AÑO",
                                "ajuste_efectivo_al": fecha(ym_date),
                                "periodo": concepto['_periodo_mes_ajuste'].split(".")[0].upper()
                            }

                            log_item(f"CARTA INCREMENTO NUMERO: {i}")
                            logger.info_yaml(data)
                            res = self.generate_pdf_template(template_name='incremento_template.md', fields=data)
                            file_name = f'{AJUSTES_DIR}/AjusteRenta_{rfc}_{concepto["CuentaPredial"]}.pdf'
                            with open(file_name, 'wb') as f:
                                f.write(res)
                            self.window.read(timeout=0)

                        if has_ajustes:
                            os.startfile(
                                os.path.abspath(AJUSTES_DIR)
                            )

                    case "recuperar_emitidas" | "recuperar_recibidas":
                        log_line("RECUPERAR")
                        fecha_final = date.today()
                        fecha_inicial = fecha_final - timedelta(days=int(values["recuperar_dias"]))
                        id_solicitud = local.config.get(event)

                        if not id_solicitud:
                            logger.info("Creando Nueva Solicitud")
                            self.window.read(timeout=0)
                            response = self.sat_service.recover_comprobante_request(
                                fecha_inicial=fecha_inicial,
                                fecha_final=fecha_final,
                                rfc_receptor=self.sat_service.signer.rfc if "recuperar_recibidas" == event else None,
                                rfc_emisor=self.sat_service.signer.rfc if "recuperar_emitidas" == event else None,
                                tipo_solicitud=TipoDescargaMasivaTerceros.CFDI,
                            )
                            logger.info_yaml(response)
                            local.config[event] = response['IdSolicitud']
                            local.config.save()
                        else:
                            logger.info_yaml({
                                'IdSolicitud': id_solicitud
                            })
                            self.window.read(timeout=0)
                            for paquete_id, data in self.recupera_comprobantes(id_solicitud):
                                if data:
                                    self.all_invoices = None
                                    with io.BytesIO(data) as b:
                                        self.unzip_cfdi(b)
                                del local.config[event]
                                local.config.save()

                    case "validate_clientes":
                        log_line("VALIDAR CLIENTES")
                        clients = ClientsManager()
                        res = PySimpleGUI.popup(
                            f"Estas seguro que quieres validar {len(clients)} clientes?",
                            title=self.window[event].ButtonText,
                            button_type=POPUP_BUTTONS_OK_CANCEL,
                        )
                        if res == "OK":
                            for rfc, details in clients.items():
                                log_item(f"VALIDANDO {rfc}")
                                self.window.read(timeout=0)
                                validar_client(rfc, details)
                            log_item("FIN")
                        else:
                            log_item("OPERACION CANCELADA")

                    case "prepare_facturas":
                        log_line("PREPARAR FACTURAS")
                        facturas = generate_ingresos(values, self.issuer, self.lugar_expedicion, self.concepto_adapter)
                        if facturas:
                            for i, cfdi in enumerate(facturas, start=1):
                                log_item(f"FACTURA NUMERO: {i}")
                                self.log_cfdi(cfdi)

                            self.invoice_button_manager.set_invoices(
                                facturas
                            )

                    case "prepare_pago":
                        log_line("COMPROBANTE PAGO")
                        if i := factura_seleccionada:
                            cfdi_pago = pago_factura(
                                factura_pagar=i,
                                fecha_pago=values["fecha_pago"],
                                forma_pago=values["forma_pago"],
                                issuer=self.issuer,
                                lugar_expedicion=self.lugar_expedicion
                            )
                            if cfdi_pago:
                                self.log_cfdi(cfdi_pago)
                                self.invoice_button_manager.set_invoices(
                                    [cfdi_pago]
                                )

                    case "ver_factura":
                        if i := factura_seleccionada:
                            os.startfile(
                                os.path.abspath(i.filename + ".pdf")
                            )

                    case "status_sat":
                        log_line("STATUS")
                        if i := factura_seleccionada:
                            estado = local_db.status_sat(i, update=True)
                            logger.info_yaml(estado)
                            local_db.describe(i)

                    case "pago_pue":
                        if i := factura_seleccionada:
                            self.log_cfdi(i)
                            local_db.describe(i)
                            st = local_db.pue_pagada(i)
                            res = PySimpleGUI.popup(
                                f"Estas seguro que quieres marcarla como {'-NO- ' if st else ''}pagada?",
                                title=self.window[event].ButtonText,
                                button_type=POPUP_BUTTONS_OK_CANCEL,
                            )
                            if res == "OK":
                                self.ch.clear()
                                local_db.pue_pagada_set(i.uuid, not st)
                                self.log_cfdi(i)
                                local_db.describe(i)
                                logger.info(f"FACTURA MARCADA COMO {'-NO- ' if st else ''}PAGADA")

                    case "ignorar_ppd":
                        if i := factura_seleccionada:
                            self.log_cfdi(i)
                            local_db.describe(i)
                            st = local_db.ppd_ignorar(i)
                            res = PySimpleGUI.popup(
                                f"Estas seguro que quieres marcarla como {'-NO- ' if st else ''}ignorada?",
                                title=self.window[event].ButtonText,
                                button_type=POPUP_BUTTONS_OK_CANCEL,
                            )
                            if res == "OK":
                                self.ch.clear()
                                local_db.ppd_ignorar_set(i.uuid, not st)
                                self.log_cfdi(i)
                                local_db.describe(i)
                                logger.info(f"FACTURA MARCADA COMO {'-NO- ' if st else ''}IGNORADA")

                    case "email_notificada":
                        if i := factura_seleccionada:
                            self.log_cfdi(i)
                            local_db.describe(i)
                            st = local_db.email_notificada(i)
                            res = PySimpleGUI.popup(
                                f"Estas seguro que quieres marcarla como {'-NO- ' if st else ''}notificada?",
                                title=self.window[event].ButtonText,
                                button_type=POPUP_BUTTONS_OK_CANCEL,
                            )
                            if res == "OK":
                                self.ch.clear()
                                local_db.email_notificada_set(i.uuid, not st)
                                self.log_cfdi(i)
                                local_db.describe(i)
                                logger.info(f"FACTURA MARCADA COMO {'-NO- ' if st else ''}NOTIFICADA")

                    case "crear_facturas":
                        log_line("CREAR FACTURAS")
                        res = PySimpleGUI.popup(
                            f"Estas seguro que quieres crear {len(invoices_to_create)} facturas?",
                            title=self.window[event].ButtonText,
                            button_type=POPUP_BUTTONS_OK_CANCEL,
                        )
                        if res == "OK":
                            self.all_invoices = None
                            for invoice in invoices_to_create:
                                try:
                                    cfdi = self.generate_invoice(invoice=invoice)
                                    logger.info(f'Factura Generada: {cfdi_header(cfdi)}')
                                    self.window.read(timeout=0)
                                except ResponseError as ex:
                                    logger.error(f'Error Generando: {cfdi_header(MyCFDI(invoice))}')
                                    logger.error(f"Status Code: {ex.response.status_code}")
                                    logger.info_yaml(ex.response.json())
                                    break
                            log_item("FIN")
                        else:
                            log_item("OPERACION CANCELADA")

                    case "prepare_correos":
                        log_line("PREPARAR CORREOS")
                        now = date.today()
                        dp = DatePeriod(now.year, now.month)
                        clients = ClientsManager()
                        a_invoices = self.get_all_invoices()

                        cfdi_correos = []
                        for receptor_rfc, notify_invoices in itertools.groupby(
                                sorted(
                                    (i for i in a_invoices.values() if not i.notified),
                                    key=lambda r: r["Receptor"]["Rfc"]
                                ),
                                lambda r: r["Receptor"]["Rfc"]
                        ):
                            notify_invoices = list(notify_invoices)
                            fac_pen = filter_invoices_iter(
                                invoices=a_invoices.values(),
                                fecha=lambda x: x < dp,
                                invoice_type="I",
                                pending_balance=lambda x: x > 0,
                                rfc_emisor=self.issuer.rfc,
                                rfc_receptor=receptor_rfc,
                                estatus='1'
                            )
                            receptor = clients[receptor_rfc]
                            cfdi_correos.append((receptor, notify_invoices, list(fac_pen)))

                        if cfdi_correos:
                            self.email_button_manager.set_emails(
                                cfdi_correos
                            )
                        else:
                            logger.info("No hay correos pendientes de enviar")

                    case "enviar_correos":
                        log_line("ENVIAR CORREOS")
                        res = PySimpleGUI.popup(
                            f"Estas seguro que quieres enviar {len(emails_to_send)} correos?",
                            title=self.window[event].ButtonText,
                            button_type=POPUP_BUTTONS_OK_CANCEL
                        )
                        if res == "OK":
                            self.enviar_correos(emails_to_send)
                            log_item("FIN")
                        else:
                            log_item("OPERACION CANCELADA")

                    case "facturas_pendientes":
                        clients = ClientsManager()
                        log_line("FACTURAS PENDIENTES")

                        fac_pen = filter_invoices_iter(
                            invoices=self.get_all_invoices().values(),
                            invoice_type="I",
                            pending_balance=lambda x: x > 0,
                            rfc_emisor=self.issuer.rfc,
                            estatus='1'
                        )
                        for receptor_rfc, fac_pen in itertools.groupby(
                                sorted(
                                    fac_pen,
                                    key=lambda r: r["Receptor"]["Rfc"]
                                ),
                                lambda r: r["Receptor"]["Rfc"]
                        ):
                            f = {
                                'Receptor': Code(receptor_rfc, clients[receptor_rfc]['RazonSocial']),
                                'FacturasPendientes': [
                                    {
                                        "Factura": f"{i.name} - {i.uuid}",
                                        'SaldoPendiente': i.saldo_pendiente,
                                        'Fecha': i["Fecha"]
                                    }
                                    for i in fac_pen
                                ]
                            }
                            logger.info_yaml(f)

                    case "descarga":
                        log_line('DESCARGADA')
                        try:
                            res = self.pac_service.recover(values["factura_pagar"], accept=Accept.XML_PDF)
                            self.all_invoices = None
                            cfdi = move_to_folder(res.xml, pdf_data=res.pdf)
                            self.log_cfdi(cfdi)
                        except DocumentNotFoundError:
                            logger.info("Factura no encontrada")

                    case "exportar_facturas":
                        log_line("EXPORTAR FACTURAS")
                        exportar_facturas(self.get_all_invoices(), values["periodo"], self.issuer)

                    case "ver_excel":
                        archivo_excel = exportar_facturas_filename(values["periodo"])
                        os.startfile(
                            os.path.abspath(archivo_excel)
                        )

                    case "periodo" | "inicio" | "final" | "fecha_pago" | "forma_pago":
                        pass

                    case _:
                        logger.error(f"Unknown event '{event}'")

            except Exception as ex:
                log_line("ERROR")
                logger.exception(ex)



import logging
import re
import warnings

# noinspection PyUnresolvedReferences
from bs4.builder import XMLParsedAsHTMLWarning
from satcfdi import csf, RFC, RFCType
from satcfdi.pacs.sat import SAT

sat_service = SAT()
logger = logging.getLogger()
warnings.filterwarnings("ignore", category=XMLParsedAsHTMLWarning)

EMAIL_REGEX = r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b'


def validar_client(rfc, details):
    try:
        rfc = RFC(rfc)
        if not rfc.is_valid():
            raise ValueError("RFC Not Valid Regex")
    except ValueError as ex:
        logger.error(f"{rfc}: {ex}")
        return

    for email in details["Email"]:
        match = re.fullmatch(EMAIL_REGEX, email)
        if not match:
            logger.error(f"{rfc}: Correo '{email}' is invalid")

    if id_cif := details["IdCIF"]:
        try:
            res = csf.retrieve(rfc, id_cif=id_cif)
        except ValueError as ex:
            logger.error(f"{rfc}: idCIF '{id_cif}' is invalid")
            return

        if rfc.type == RFCType.MORAL:
            if details['RazonSocial'] != res['Denominación o Razón Social']:
                logger.error(f"{rfc}: RazonSocial '{details['RazonSocial']}' is invalid, expected '{res['Denominación o Razón Social']}'")
        elif rfc.type == RFCType.FISICA:
            if details['RazonSocial'] != f"{res['Nombre']} {res['Apellido Paterno']} {res['Apellido Materno']}":
                logger.error(f"{rfc}: RazonSocial '{details['RazonSocial']}' is invalid, expected '{res['Nombre']} {res['Apellido Paterno']} {res['Apellido Materno']}'")

        if details['CodigoPostal'] != res['CP']:
            logger.error(f"{rfc}: CodigoPostal '{details['CodigoPostal']}' is invalid, expected '{res['CP']}'")

        if details['RegimenFiscal'] not in (r['RegimenFiscal'] for r in res['Regimenes']):
            logger.error(
                f"{rfc}: RegimenFiscal '{details['RegimenFiscal']}' is invalid, "
                f"expected '{(r['RegimenFiscal'].code for r in res['Regimenes'])}'"
            )

        if res['Situación del contribuyente'] not in ['ACTIVO', 'REACTIVADO']:
            logger.error(f"{rfc}: Is not ACTIVO '{res['Situación del contribuyente']}'")

        taxpayer_status = sat_service.list_69b(rfc)
        if taxpayer_status:
            logger.error(f"{rfc}: has status '{taxpayer_status}'")


def clientes_generar_txt(clients):
    with open("clientes.txt", 'w') as f:
        for i, (cliente_rfc, cliente_data) in enumerate(clients.items(), start=1):
            f.write(f"{i}|{cliente_rfc}|{cliente_data['RazonSocial']}|{cliente_data['CodigoPostal']}")
            f.write("\n")



# from https://github.com/AzureAD/microsoft-authentication-library-for-python
# modified to not use threading so as to wait for response
import logging
import sys
from http.server import HTTPServer, BaseHTTPRequestHandler
from string import Template
from urllib.parse import urlparse, parse_qs

logger = logging.getLogger()


def _qs2kv(qs):
    """Flatten parse_qs()'s single-item lists into the item itself"""
    return {k: v[0] if isinstance(v, list) and len(v) == 1 else v
            for k, v in qs.items()}


# success_template = Template("Authentication completed. You can close this window now.")
error_template = Template("Authentication failed. $error: $error_description. ($error_uri)")
welcome_template = "Welcome"

success_template = Template('<html><body>Authentication completed. You can close this window now.<script>setTimeout("window.close()",3000)</script></body></html>')


class _AuthCodeHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        # For flexibility, we choose to not check self.path matching redirect_uri
        # assert self.path.startswith('/THE_PATH_REGISTERED_BY_THE_APP')
        qs = parse_qs(urlparse(self.path).query)
        if qs.get('code') or qs.get("error"):  # So, it is an auth response
            self.server.auth_response = _qs2kv(qs)
            logger.debug("Got auth response: %s", self.server.auth_response)
            template = (success_template
                        if "code" in qs else error_template)
            self._send_full_response(
                template.safe_substitute(**self.server.auth_response))
            # NOTE: Don't do self.server.shutdown() here. It'll halt the server.
        else:
            self._send_full_response(welcome_template)

    def _send_full_response(self, body, is_ok=True):
        self.send_response(200 if is_ok else 400)
        content_type = 'text/plain'
        content_type = 'text/html'
        self.send_header('Content-type', content_type)
        self.end_headers()
        self.wfile.write(body.encode("utf-8"))

    def log_message(self, format, *args):
        logger.debug(format, *args)  # To override the default log-to-stderr behavior


def is_wsl():
    # "Official" way of detecting WSL: https://github.com/Microsoft/WSL/issues/423#issuecomment-221627364
    # Run `uname -a` to get 'release' without python
    #   - WSL 1: '4.4.0-19041-Microsoft'
    #   - WSL 2: '4.19.128-microsoft-standard'
    import platform
    uname = platform.uname()
    platform_name = getattr(uname, 'system', uname[0]).lower()
    release = getattr(uname, 'release', uname[2]).lower()
    return platform_name == 'linux' and 'microsoft' in release


class _AuthCodeHttpServer(HTTPServer, object):
    # address_family = socket.AF_INET6 # if using ipv6
    def __init__(self, server_address, *args, **kwargs):
        _, port = server_address
        if port and (sys.platform == "win32" or is_wsl()):
            # The default allow_reuse_address is True. It works fine on non-Windows.
            # On Windows, it undesirably allows multiple servers listening on same port,
            # yet the second server would not receive any incoming request.
            # So, we need to turn it off.
            self.allow_reuse_address = False
        super(_AuthCodeHttpServer, self).__init__(server_address, *args, **kwargs)

    def handle_timeout(self):
        # It will be triggered when no request comes in self.timeout seconds.
        # See https://docs.python.org/3/library/socketserver.html#socketserver.BaseServer.handle_timeout
        raise RuntimeError("Timeout. No auth response arrived.")  # Terminates this server
        # We choose to not call self.server_close() here,
        # because it would cause a socket.error exception in handle_request(),
        # and likely end up the server being server_close() twice.


class AuthCodeReceiver(object):
    # This class has (rather than is) an _AuthCodeHttpServer, so it does not leak API
    def __init__(self, port=None):
        address = "127.0.0.1"
        self._server = _AuthCodeHttpServer((address, port or 0), _AuthCodeHandler)
        self._closing = False

    def get_port(self):
        """The port this server actually listening to"""
        # https://docs.python.org/2.7/library/socketserver.html#SocketServer.BaseServer.server_address
        return self._server.server_address[1]

    def get_auth_response(self, timeout=None, state=None):
        self._server.timeout = timeout  # Otherwise its handle_timeout() won't work
        self._server.auth_response = {}  # Shared with _AuthCodeHandler
        while not self._closing:  # Otherwise, the handle_request() attempt
            self._server.handle_request()
            if self._server.auth_response:
                if state and state != self._server.auth_response.get("state"):
                    logger.debug("State mismatch. Ignoring this noise.")
                else:
                    break

        response = self._server.auth_response
        self._server.server_close()
        return response or None  # Return via writable result param

    def close(self):
        """Either call this eventually; or use the entire class as context manager"""
        self._closing = True
        self._server.server_close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



import logging

import yaml
from satcfdi import Code


class LogAdapter(logging.LoggerAdapter):
    def info_yaml(self, data):
        self.info(
            yaml.safe_dump(data, allow_unicode=True, width=1280, sort_keys=False)
        )


logger = LogAdapter(logging.getLogger())


class LogHandler(logging.StreamHandler):
    def __init__(self, console):
        super().__init__()
        self.buffer = ""
        self.console = console

    def emit(self, record):
        msg = self.format(record)
        self.buffer += msg + '\n'
        self.console.update(value=self.buffer)

    def clear(self):
        self.buffer = ""
        self.console.update(value="")


def log_email(receptor, notify_invoices, facturas_pendientes):
    logger.info_yaml({
        "Rfc": Code(receptor["Rfc"], receptor["RazonSocial"]),
        "Facturas": [f"{i.name} - {i.uuid}" for i in notify_invoices],
        "PendientesMesesAnteriores": [f"{i.name} - {i.uuid}" for i in facturas_pendientes],
        "Correos": receptor["Email"]
    })


def log_line(text, exc_info=False):
    ln = (150 - len(text)) // 2
    logger.info(
        ("=" * ln) + " " + text + " " + ("=" * ln),
        exc_info=exc_info
    )


def log_item(text, exc_info=False):
    ln = (150 - len(text)) // 2
    logger.info(
        ("*" * ln) + " " + text + " " + ("*" * ln),
        exc_info=exc_info
    )


def cfdi_header(cfdi):
    receptor = Code(cfdi['Receptor']['Rfc'], cfdi['Receptor']['Nombre'])
    return f"{cfdi.name} - {cfdi.uuid} {receptor}"



from importlib.metadata import version, PackageNotFoundError

__title__ = "SAT-DIGITAL-INVOICE"
__package__ = "satdigitalinvoice"

__description__ = "GUI APP to Generate CFDI"
__git_project__ = "SAT-CFDI/python-satdigitalinvoice"
__url__ = f"https://github.com/{__git_project__}"

__docs_url__ = f"https://{__package__}.readthedocs.io"
__docs_badge_url__ = f"https://readthedocs.org/projects/{__package__}/badge"

__author__ = "satcfdi@outlook.com"
__author_email__ = "satcfdi@outlook.com"
__license__ = "MIT License"

try:
    __version__ = version(__package__)
except PackageNotFoundError:
    __version__ = "1.0.0"

__user_agent__ = f"python-{__package__}/{__version__}"



import hashlib
import posixpath
import secrets
import webbrowser
from base64 import urlsafe_b64encode
from random import randrange

import requests

from .auth_code_receiver import AuthCodeReceiver
from .jwt_util import load_jwt


def get_token(
        issuer_url,
        app_client_id,
        redirect_url=None,
        scopes=None,
        origin=None,
        login_hint=None,
        domain_hint=None,
        browser_name=None
):
    state = str(randrange(1000000000))  # "InfoOfWhereTheUserWantedToGo"
    nonce = str(randrange(1000000000))

    # Your first step is to generate_key_pair a code verifier and challenge:
    # Code verifier: Random URL-safe string with a minimum length of 43 characters.
    # Code challenge: Base64 URL-encoded SHA-256 hash of the code verifier.
    code_verifier = secrets.token_urlsafe(43)
    m = hashlib.sha256()
    m.update(code_verifier.encode())
    code_challenge = urlsafe_b64encode(m.digest()).decode().rstrip("=")

    if redirect_url is None:
        server = AuthCodeReceiver(port=0)
        port = server.get_port()
        redirect_url = f"http://localhost:{port}/"
    else:
        server = AuthCodeReceiver(port=int(redirect_url.split(":")[-1]))

    res = requests.get(
        url=posixpath.join(issuer_url, "oauth2/v2.0/authorize"),
        params={
            "client_id": app_client_id,
            "response_type": "code",
            "scope": " ".join(scopes or []),
            "redirect_uri": redirect_url,
            "state": state,
            "nonce": nonce,  # Optional
            "code_challenge_method": "S256",
            "code_challenge": code_challenge,
            "client_info": "1",
            "prompt": "select_account",
            "login_hint": login_hint or "",
            "domain_hint": domain_hint or ""
        }
    )

    assert res.status_code == 200

    if browser_name:
        webbrowser.get(browser_name).open(res.request.url)
    else:
        webbrowser.open(res.request.url)

    result = server.get_auth_response(timeout=1000, state=state)
    q = result

    token = requests.post(
        url=posixpath.join(issuer_url, "oauth2/v2.0/token"),
        data={
            "grant_type": "authorization_code",
            "client_id": app_client_id,
            "code_verifier": code_verifier,
            "redirect_uri": redirect_url,
            "code": q["code"],
        },
        headers={
            'content-type': 'application/x-www-form-urlencoded',
            'origin': origin
        },
        # auth=HTTPBasicAuth(CLIENT_ID, CLIENT_SECRET),
    )

    if token.status_code == 200:
        return token.json()

    assert token.status_code == 200


def token_refresh(
        token,
        issuer_url=None,
        scopes=None,
        origin=None
):
    header, payload, _ = load_jwt(token['access_token'])

    issuer_url = issuer_url or payload["iss"]
    app_client_id = payload["appid"]  # Azure portal

    data = {
        "grant_type": "refresh_token",
        "client_id": app_client_id,
        "refresh_token": token['refresh_token']
    }
    if scopes:
        data["scope"] = " ".join(scopes or []),

    token = requests.post(
        url=posixpath.join(issuer_url, "oauth2/v2.0/token"),
        data=data,
        headers={
            'content-type': 'application/x-www-form-urlencoded',
            'origin': origin
        },
        # auth=HTTPBasicAuth(CLIENT_ID, CLIENT_SECRET),
    )

    if token.status_code == 200:
        return token.json()

    raise Exception(token.json())




import glob
import logging
import os
from collections.abc import Mapping
from datetime import datetime
from decimal import Decimal
from typing import MutableMapping
from uuid import UUID

from satcfdi.accounting import complement_invoices_data, SatCFDI

from .local import local_db, load_data, save_data

ALL_INVOICES = 'all_invoices'
ALL_RETENCIONES = 'all_retenciones'
logger = logging.getLogger()

PPD = "PPD"
PUE = "PUE"


class MyCFDI(SatCFDI):
    issuer_rfc = None

    def consulta_estado(self):
        return local_db.status_sat(self)

    @SatCFDI.estatus.getter
    def estatus(self) -> str:
        Estatus = {
            "Cancelado": "0",
            "Vigente": "1"
        }
        estado = local_db.status_sat(self).get('Estado')
        return Estatus.get(estado, "1")

    @SatCFDI.fecha_cancelacion.getter
    def fecha_cancelacion(self) -> datetime:
        return None

    @SatCFDI.saldo_pendiente.getter
    def saldo_pendiente(self) -> Decimal:
        if self['TipoDeComprobante'] == "I":
            if self["MetodoPago"] == PPD:
                if local_db.ppd_ignorar(self):
                    return Decimal()

            if self["MetodoPago"] == PUE and not self.payments and self["Emisor"]["Rfc"] == self.issuer_rfc:
                if local_db.pue_pagada(self):
                    return Decimal()
                else:
                    return self['Total']

        return super().saldo_pendiente

    @classmethod
    def get_all_invoices(cls, invoices: MutableMapping, search_path="*.xml") -> bool:
        # Check that all names are correct
        for file in glob.iglob(search_path, recursive=True):
            if not cls.uuid_from_filename(filename=file):
                invoice = cls.from_file(file)
                rename_invoice(file, invoice)

        # Load Invoices
        dup_check = set()
        was_updated = False

        for file in glob.iglob(search_path, recursive=True):
            fid = cls.uuid_from_filename(filename=file)
            if fid:
                if fid not in invoices:
                    was_updated = True
                    invoices[fid] = cls.from_file(file)
            else:
                raise Exception("CFDI with invalid File Name")

            # Check we don't have a duplicate
            if fid not in dup_check:
                dup_check.add(fid)
            else:
                raise Exception("Duplicated Invoice Found", fid, file)

        # Remove extra
        if len(dup_check) < len(invoices):
            was_updated = True
            for i in list(invoices.keys()):
                if i not in dup_check:
                    del invoices[i]

        # if the length does not match, then we need to try again :-/
        return was_updated

    @property
    def filename(self):
        match self.tag:
            case '{http://www.sat.gob.mx/cfd/3}Comprobante' | '{http://www.sat.gob.mx/cfd/4}Comprobante':
                return "facturas/{3:%Y}/{3:%Y-%m}/{4}/{3:%Y-%m-%d}_{0}_[{1}]_{2}".format(
                    self.name,
                    self["TipoDeComprobante"].code,
                    self.uuid,
                    self["Fecha"],
                    "Emitidas" if self["Emisor"]["Rfc"] == self.issuer_rfc else "Recibidas"
                )
            case '{http://www.sat.gob.mx/esquemas/retencionpago/1}Retenciones' | '{http://www.sat.gob.mx/esquemas/retencionpago/2}Retenciones':
                return "retenciones/{2:%Y}/{2:%Y-%m}/{3}/{2:%Y-%m-%d}_{0}_{1}".format(
                    self.get("FolioInt", ""),
                    self.uuid,
                    self["FechaExp"],
                    "Emitidas" if self["Emisor"]["RFCEmisor"] == self.issuer_rfc else "Recibidas"
                )

    @classmethod
    def uuid_from_filename(cls, filename):
        filename = os.path.basename(filename)
        parts = os.path.splitext(filename)[0].split("_")
        if len(parts) >= 3:
            try:
                return UUID(parts[-1])
            except ValueError:
                pass
        return None

    @property
    def notified(self) -> bool:
        return self["Emisor"]["Rfc"] != self.issuer_rfc \
            or local_db.email_notificada(self)

    @notified.setter
    def notified(self, value):
        local_db.email_notificada_set(self.uuid, value)


def rename_invoice(file, invoice: MyCFDI, include_pdf=True):
    preferred_filename = invoice.filename
    if preferred_filename + ".xml" != file:
        os.makedirs(os.path.dirname(preferred_filename), exist_ok=True)

        try:
            os.rename(file, preferred_filename + ".xml")
        except FileExistsError:
            os.remove(file)

        if include_pdf:
            try:
                os.rename(file[:-4] + ".pdf", preferred_filename + ".pdf")
            except FileExistsError:
                os.remove(file[:-4] + ".pdf")
            except FileNotFoundError:
                logger.warning("PDF Not found for %s, will create one instead", preferred_filename + ".pdf")
                invoice.pdf_write(preferred_filename + ".pdf")


def move_to_folder(xml_data, pdf_data):
    cfdi = MyCFDI.from_string(xml_data)

    full_name = cfdi.filename
    os.makedirs(os.path.dirname(full_name), exist_ok=True)

    try:
        with open(full_name + ".xml", 'xb') as fp:
            fp.write(xml_data)
        logger.info("Factura ha sido agregada: '%s'", full_name)

        if pdf_data:
            with open(full_name + ".pdf", 'wb') as fp:
                fp.write(pdf_data)
        else:
            try:
                cfdi.pdf_write(full_name + ".pdf")
            except:
                logger.exception("Fallo crear PDF: '%s'", full_name)
    except FileExistsError:
        logger.info("Factura ya se tenia: %s", full_name)

    return cfdi


def get_all_cfdi() -> Mapping[UUID, MyCFDI]:
    all_invoices = load_data(ALL_INVOICES, {})

    has_updates = MyCFDI.get_all_invoices(invoices=all_invoices, search_path="facturas/**/*.xml")
    if has_updates:
        save_data(ALL_INVOICES, all_invoices)

    complement_invoices_data(all_invoices)
    return all_invoices


def get_all_retenciones() -> Mapping[UUID, MyCFDI]:
    all_invoices = load_data(ALL_RETENCIONES, {})

    has_updates = MyCFDI.get_all_invoices(invoices=all_invoices, search_path="retenciones/**/*.xml")
    if has_updates:
        save_data(ALL_RETENCIONES, all_invoices)

    return all_invoices



import base64
import hashlib
import hmac
import json
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, padding
from hmac import HMAC


def create_signature(data, key, header):
    alg = header["alg"]

    match alg[:2]:
        case "HS":
            algorithm = {
                'HS256': hashlib.sha256,
                'HS384': hashlib.sha384,
                'HS512': hashlib.sha512
            }[alg]

            return HMAC(key, data, algorithm).digest()

        case 'RS' | 'PS':
            algorithm = {
                'RS256': hashes.SHA256,
                'RS384': hashes.SHA384,
                'RS512': hashes.SHA512,
                'PS256': hashes.SHA256,
                'PS384': hashes.SHA384,
                'PS512': hashes.SHA512,
            }[alg]()

            if alg.startswith("RS"):
                pad = padding.PKCS1v15()
            else:
                pad = padding.PSS(
                    mgf=padding.MGF1(algorithm),
                    salt_length=algorithm.digest_size
                )

            return key.sign(
                data=data,
                padding=pad,
                algorithm=algorithm
            )


def make_jwt(key, header, payload):
    header_segment = _base64url_encode(json.dumps(header).encode())
    payload_segment = _base64url_encode(json.dumps(payload).encode())

    signed_data = (header_segment + "." + payload_segment).encode()
    signature = create_signature(signed_data, key, header)
    crypto_segment = _base64url_encode(signature)

    return header_segment + "." + payload_segment + "." + crypto_segment


def load_jwt(jwt):
    try:
        header_segment, payload_segment, crypto_segment = jwt.split('.', 2)

        header = json.loads(_base64url_decode(header_segment))
        payload = json.loads(_base64url_decode(payload_segment))
        signature = _base64url_decode(crypto_segment)

        return header, payload, signature
    except Exception as e:  # (ValueError, TypeError)
        raise Exception("Unable to parse token") from e


def _base64url_decode(input):
    input += '=' * (-len(input) % 4)
    return base64.urlsafe_b64decode(input)


def _base64url_encode(input):
    input = base64.urlsafe_b64encode(input).decode()
    return input.rstrip("=")



import logging
from datetime import datetime, date

import xlsxwriter
from babel.dates import format_date
from satcfdi import DatePeriod
from satcfdi.accounting import filter_invoices_iter, filter_payments_iter, invoices_export, payments_export
from satcfdi.accounting.process import IVA16
from satcfdi.create.cfd import cfdi40
from satcfdi.create.cfd.cfdi40 import Comprobante
from satcfdi.pacs import sat
# noinspection PyUnresolvedReferences
from satcfdi.transform.catalog import CATALOGS
from xlsxwriter.exceptions import FileCreateError

from .file_data_managers import FacturasManager, ClientsManager, environment_default
from .local import config

logger = logging.getLogger()
sat_manager = sat.SAT()

PERIODOS = {
    "Mensual": 1,
    "Bimestral": 2,
    "Trimestral": 3,
    "Cuatrimestral": 4,
    "Semestral": 6,
    "Anual": 12
}


def create_cfdi(client, factura_details, serie, folio, issuer, lugar_expedicion):
    invoice = cfdi40.Comprobante(
        emisor=issuer,
        lugar_expedicion=lugar_expedicion,
        receptor=cfdi40.Receptor(
            rfc=factura_details['Rfc'],
            nombre=client['RazonSocial'],
            uso_cfdi=factura_details['UsoCFDI'],
            domicilio_fiscal_receptor=client['CodigoPostal'],
            regimen_fiscal_receptor=client['RegimenFiscal']
        ),
        metodo_pago=factura_details['MetodoPago'],
        forma_pago=factura_details['FormaPago'],
        serie=serie,
        folio=folio,
        conceptos=factura_details["Conceptos"]
    ).process()

    if factura_details['Total'] != invoice['Total']:
        logger.error(f"{factura_details['Rfc']}: Total '{factura_details['Total']}' is invalid, expected '{invoice['Total']}'")
        return None

    return invoice


def parse_periodo_mes_ajuste(periodo_mes_ajuste: str):
    parts = periodo_mes_ajuste.split(".")
    if len(parts) != 2:
        raise ValueError("Periodo Invalido")

    periodo, mes_ajuste = parts
    if not mes_ajuste.isnumeric():
        raise ValueError("Periodo Invalido")

    mes_ajuste = int(mes_ajuste)
    if not (12 >= int(mes_ajuste) >= 1):
        raise ValueError("Periodo Invalido")

    if periodo not in PERIODOS:
        raise ValueError("Periodo Invalido")

    return periodo, mes_ajuste


def parse_ym_date(periodo):
    return datetime.strptime(periodo, '%Y-%m')


def parse_date_period(periodo):
    if "-" in periodo:
        period = parse_ym_date(periodo)
        return DatePeriod(period.year, period.month)
    else:
        return DatePeriod(int(periodo))


def format_concepto_desc(concepto, periodo):
    template = environment_default.from_string(concepto["Descripcion"])
    concepto["Descripcion"] = template.render(
        periodo=periodo
    )


def validad_facturas(clients, facturas):
    is_valid = True
    for factura_details in facturas:
        cliente = clients.get(factura_details['Rfc'])
        if not cliente:
            logger.error(f"{factura_details['Rfc']}: client not found")
            is_valid = False

        if conceptos := factura_details.get("Conceptos"):
            for c in conceptos:
                periodo_mes_ajuste = c.get("_periodo_mes_ajuste", "")
                try:
                    parse_periodo_mes_ajuste(periodo_mes_ajuste)
                except ValueError:
                    logger.error(f"{factura_details['Rfc']}: _periodo_mes_ajuste '{periodo_mes_ajuste}' is invalid")
                    is_valid = False
        else:
            logger.error(f"{factura_details['Rfc']}: no tienen conceptos")

        if factura_details["MetodoPago"] == "PPD" and factura_details["FormaPago"] != "99":
            logger.error(f"{factura_details['Rfc']}: FormaPago '{factura_details['FormaPago']}' is invalid, expected '99' for PPD")
            is_valid = False

    return is_valid


def year_month_desc(year, month):
    return format_date(date(year=year, month=month, day=1), locale='es_MX', format="'Mes de' MMMM 'del' y").upper()


def perpare_concepto(ym_date, concepto, concepto_adapter):
    periodo, mes_ajuste = parse_periodo_mes_ajuste(concepto['_periodo_mes_ajuste'])
    periodo_meses = PERIODOS[periodo]

    if (ym_date.month - mes_ajuste) % periodo_meses == 0:
        periodo = year_month_desc(ym_date.year, ym_date.month)
        if periodo_meses > 1:
            mes_final = (ym_date.month + periodo_meses - 2) % 12 + 1
            periodo += " AL " + year_month_desc(
                year=ym_date.year + int(mes_final < ym_date.month),
                month=mes_final
            )

        format_concepto_desc(concepto, periodo=periodo)
        if concepto_adapter:
            concepto_adapter(concepto)
        return concepto
    return None


def generate_ingresos(values, issuer, lugar_expedicion, concepto_adapter):
    clients = ClientsManager()
    facturas = FacturasManager()["Facturas"]
    if not validad_facturas(clients, facturas):
        return

    inicio = int(values["inicio"])
    if inicio <= 0:
        logger.error("Inicio no Valido")

    final = values["final"] or len(facturas)
    final = int(final)
    if final <= 0 or final < inicio:
        logger.error("Final no Valido")

    ym_date = parse_ym_date(values["periodo"])
    facturas = facturas[inicio - 1:final]
    folio = config.folio()
    serie = config.serie()

    def facturas_iter():
        i = 0
        for f in facturas:
            client = clients[f['Rfc']]
            conceptos = [x for x in (perpare_concepto(ym_date, c, lambda x: concepto_adapter(client, x)) for c in f["Conceptos"]) if x]
            if conceptos:
                f["Conceptos"] = conceptos
                yield create_cfdi(client, f, serie, str(folio + i), issuer, lugar_expedicion)
                i += 1

    cfdis = [i for i in facturas_iter()]

    if None in cfdis:
        return

    return cfdis


def parse_fecha_pago(fecha_pago):
    if not fecha_pago:
        logger.error("Fecha de Pago esta vacia")
        return

    fecha_pago = datetime.fromisoformat(fecha_pago)
    if fecha_pago > datetime.now():
        logger.error("Fecha de Pago esta en el futuro")
        return

    if fecha_pago.replace(hour=12) > datetime.now():
        fecha_pago = datetime.now()
    else:
        fecha_pago = fecha_pago.replace(hour=12)

    dif = datetime.now() - fecha_pago
    if dif.days > 30:
        logger.error("!!! FECHA DE PAGO ES MAYOR A 30 DIAS !!!")

    return fecha_pago


def pago_factura(factura_pagar, fecha_pago, forma_pago, issuer, lugar_expedicion):
    if not (fecha_pago := parse_fecha_pago(fecha_pago)):
        return

    i = factura_pagar
    if i.saldo_pendiente != i["Total"]:
        logger.error("Comprobante ya tiene pago anterior")
        logger.error(f"Saldo Pendiente: {i.saldo_pendiente}")
        return

    return Comprobante.pago_comprobantes(
        emisor=issuer,
        lugar_expedicion=lugar_expedicion,
        comprobantes=i,
        fecha_pago=fecha_pago,
        forma_pago=forma_pago,
        serie=config.serie(),
        folio=str(config.folio()),
    ).process()

    # Pago Parcial
    # comprobantes = [PagoComprobante(
    #     comprobante=c,
    #     num_parcialidad=c.ultima_num_parcialidad + 1,
    #     imp_saldo_ant=c.saldo_pendiente,
    #     imp_pagado=Decimal(imp_pagado)
    # ) for c, imp_pagado in cfdis],


def find_ajustes(mes_ajuste):
    for f in FacturasManager()["Facturas"]:
        rfc = f["Rfc"]
        for concepto in f["Conceptos"]:
            _, mes_aj = parse_periodo_mes_ajuste(concepto['_periodo_mes_ajuste'])
            if mes_aj == mes_ajuste:
                yield rfc, concepto


def exportar_facturas_filename(periodo):
    dp = parse_date_period(periodo)

    if dp.month:
        archivo_excel = f"facturas/{dp.year}/{dp}/{dp}.xlsx"
    else:
        archivo_excel = f"facturas/{dp.year}/{dp}.xlsx"

    return archivo_excel


def exportar_facturas(all_invoices, periodo, issuer):
    dp = parse_date_period(periodo)

    emitidas = filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_emisor=issuer.rfc)
    emitidas_pendientes = filter_invoices_iter(
        invoices=all_invoices.values(), fecha=lambda x: x <= dp, rfc_emisor=issuer.rfc, estatus='1',
        invoice_type="I", pending_balance=lambda x: x > 0)
    emitidas_pagos = filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_emisor=issuer.rfc)

    recibidas = filter_invoices_iter(invoices=all_invoices.values(), fecha=dp, rfc_receptor=issuer.rfc)
    recibidas_pendientes = filter_invoices_iter(
        invoices=all_invoices.values(), fecha=lambda x: x <= dp, rfc_receptor=issuer.rfc, estatus='1',
        invoice_type="I", pending_balance=lambda x: x > 0)
    recibidas_pagos = filter_payments_iter(invoices=all_invoices, fecha=dp, rfc_receptor=issuer.rfc)

    recibidas_pagos = list(recibidas_pagos)
    pagos_hechos_iva = [
        p
        for p in recibidas_pagos
        if p.impuestos.get("Traslados", {}).get(IVA16, {}).get("Importe", 0) > 0
           and p.comprobante["Receptor"].get("RegimenFiscalReceptor") in (issuer.tax_system, None)
    ]
    prediales = [p for p in recibidas_pagos if p.comprobante["Emisor"]["Rfc"] == "TMT620101EV4"]

    archivo_excel = exportar_facturas_filename(periodo)
    workbook = xlsxwriter.Workbook(archivo_excel)
    # EMITIDAS
    invoices_export(workbook, "EMITIDAS", emitidas)
    invoices_export(workbook, "EMITIDAS PENDIENTES", emitidas_pendientes)
    payments_export(workbook, "EMITIDAS PAGOS", emitidas_pagos)

    # RECIBIDAS
    invoices_export(workbook, "RECIBIDAS", recibidas)
    invoices_export(workbook, "RECIBIDAS PENDIENTES", recibidas_pendientes)
    payments_export(workbook, "RECIBIDAS PAGOS", recibidas_pagos)

    # SPECIALES
    payments_export(workbook, f"RECIBIDAS PAGOS IVA {issuer.tax_system.code}", pagos_hechos_iva)
    if prediales:
        payments_export(workbook, "PREDIALES", prediales)

    try:
        workbook.close()
        logger.info(f"Archivo {archivo_excel} creado")
    except FileCreateError:
        logger.info(f"No se pudo crear el archivo {archivo_excel}")



from satcfdi import Signer
from satcfdi.accounting.email import EmailManager
from satcfdi.pacs import Environment
from satcfdi.pacs.diverza import Diverza

from satdigitalinvoice.__main__ import FacturacionGUI
from satdigitalinvoice.layout import make_layout


def test_main():
    pac_service = Diverza(
        rfc="asdfas",
        id='123',
        token='asdf',
        environment=Environment.TEST
    )

    fiel_signer = None

    csd_signer = Signer.load(
        certificate=open('csd/cacx7605101p8.cer', 'rb').read(),
        key=open('csd/cacx7605101p8.key', 'rb').read(),
        password=open('csd/cacx7605101p8.txt', 'rb').read(),
    )

    email_manager = EmailManager(
        stmp_host="hola.com",
        stmp_port=123,
        imap_host="hola.com",
        imap_port=789,
        user="someuser@email.com",
        password="1234"
    )

    app = FacturacionGUI(
        pac_service=pac_service,
        csd_signer=csd_signer,
        fiel_signer=fiel_signer,
        email_manager=email_manager,
    )


# function to check the uniqueness of keys in pysimplegui layout
def check_unique_keys(layout):
    keys = []
    for element in layout:
        if isinstance(element, list):
            check_unique_keys(element)
        else:
            if element[0] in keys:
                raise Exception(f"Key {element[0]} is not unique")
            else:
                keys.append(element[0])


def test_layout_unique_keys():
    layout = make_layout(has_fiel=True)

    def elements(layout):
        for e in layout:
            if isinstance(e, list):
                yield from elements(e)
            else:
                yield e

    unique_keys = set()
    for e in elements(layout):
        if e.Key:
            if e.Key in unique_keys:
                raise Exception(f"Key {e.Key} is not unique")
            else:
                unique_keys.add(e.Key)



from babel.dates import format_date
from dateutil import rrule
from num2words import num2words


def pesos(amount):
    amount = round(amount, 2)
    decimal_part = "{:.2f}".format(amount).split(".")[1]
    integer_part = int(amount)
    integer_part = number_to_currency(integer_part).upper()

    return "${0:,.2f} (SON: {1} {2}/100M.N.)".format(amount, integer_part, decimal_part)


def porcentaje(amount, decimals=2):
    percentage = amount * 100
    percentage = round(percentage, decimals)
    percentage_text = number_to_text(percentage).upper()

    return "{0}% ({1} POR CIENTO)".format(percentage, percentage_text)


def number_to_currency(number):
    return num2words(number, lang='es_CO', to='currency')


def number_to_text(number):
    return num2words(number, lang='es_CO')


def date_range(start, end):
    return "{0} AL {1}".format(fecha(start), fecha(end))


def fecha(date):
    return format_date(date, locale='es_MX', format="d 'de' MMMM 'del' y").upper()


def data_range_length_months(start, end):
    months = len(list(rrule.rrule(rrule.MONTHLY, dtstart=start, until=end)))
    months_text = number_to_text(months).upper()
    return "{0} ({1} MESES)".format(months, months_text)


def meses(count):
    months = count
    months_text = number_to_text(months).upper()
    return "{0} ({1} MESES)".format(months, months_text)



import logging
import os
import pickle
from uuid import UUID

import diskcache
from satcfdi.accounting import SatCFDI
from satcfdi.pacs import sat
from .log_tools import LogAdapter
from .file_data_managers import ConfigManager

PUE_PAGADA = 0
PPD_IGNORAR = 1
EMAIL_NOTIFICADA = 2
STATUS_SAT = 3

config = ConfigManager()
sat_manager = sat.SAT()

logger = LogAdapter(logging.getLogger())
DATA_DIR = '.data'


class LocalDB:
    def __init__(self):
        self.local_storage = diskcache.Cache(os.path.join(DATA_DIR, 'cache'))

    def pue_pagada(self, uuid: UUID):
        return self.local_storage.get(
            (PUE_PAGADA, uuid), False
        )

    def pue_pagada_set(self, uuid: UUID, value: bool):
        if value:
            self.local_storage[(PUE_PAGADA, uuid)] = value
        else:
            try:
                del self.local_storage[(PUE_PAGADA, uuid)]
            except KeyError:
                pass

    def ppd_ignorar(self, uuid: UUID):
        return self.local_storage.get(
            (PPD_IGNORAR, uuid), False
        )

    def ppd_ignorar_set(self, uuid: UUID, value: bool):
        if value:
            self.local_storage[(PPD_IGNORAR, uuid)] = value
        else:
            try:
                del self.local_storage[(PPD_IGNORAR, uuid)]
            except KeyError:
                pass

    def email_notificada(self, uuid: UUID):
        return self.local_storage.get(
            (EMAIL_NOTIFICADA, uuid), False
        )

    def email_notificada_set(self, uuid: UUID, value: bool):
        if value:
            self.local_storage[(EMAIL_NOTIFICADA, uuid)] = value
        else:
            try:
                del self.local_storage[(EMAIL_NOTIFICADA, uuid)]
            except KeyError:
                pass

    def status_sat(self, uuid: UUID):
        return self.local_storage.get(
            (STATUS_SAT, uuid), {}
        )

    def status_sat_set(self, uuid: UUID, value: dict):
        if value:
            self.local_storage[(STATUS_SAT, uuid)] = value
        else:
            try:
                del self.local_storage[(STATUS_SAT, uuid)]
            except KeyError:
                pass


class LocalDBSatCFDI(LocalDB):
    def email_notificada(self, cfdi: SatCFDI):
        if cfdi["Fecha"] <= config['email_notificada_hasta']:
            return True
        return super().email_notificada(cfdi.uuid)

    def pue_pagada(self, cfdi: SatCFDI):
        if cfdi["Fecha"] <= config['pue_pagada_hasta']:
            return True
        return super().pue_pagada(cfdi.uuid)

    def ppd_ignorar(self, cfdi: SatCFDI):
        if cfdi["Fecha"] <= config['ppd_ignorar_hasta']:
            return True
        return super().ppd_ignorar(cfdi.uuid)

    def status_sat(self, cfdi: SatCFDI, update=False):
        if update:
            res = sat_manager.status(cfdi)
            if res["ValidacionEFOS"] == "200":
                self.status_sat_set(cfdi.uuid, res)
            return res
        else:
            return super().status_sat(cfdi.uuid)

    def describe(self, cfdi: SatCFDI):
        logger.info_yaml({
            'pue_pagada': self.pue_pagada(cfdi),
            'ppd_ignorar': self.ppd_ignorar(cfdi),
            'email_notificada': self.email_notificada(cfdi),
            'status_sat': self.status_sat(cfdi)
        })


local_db = LocalDBSatCFDI()


def save_data(file, data):
    with open(os.path.join(DATA_DIR, file), 'wb') as f:
        pickle.dump(data, f)


def load_data(file, default=None):
    try:
        with open(os.path.join(DATA_DIR, file), 'rb') as f:
            return pickle.load(f)
    except FileNotFoundError:
        return default



from datetime import date

import PySimpleGUI as sg
from dateutil.relativedelta import relativedelta
# noinspection PyUnresolvedReferences
from satcfdi.transform.catalog import CATALOGS

from .log_tools import *

FORMA_PAGO = CATALOGS['{http://www.sat.gob.mx/sitio_internet/cfd/catalogos}c_FormaPago']
TEXT_PADDING = ((5, 0), 3)
RTEXT_PADDING = ((0, 0), 3)


def make_layout(has_fiel):
    # LAYOUT
    button_column = [
        sg.Button("Preparar Facturas", key="prepare_facturas", border_width=0),
        sg.Text("Año-Mes:", pad=TEXT_PADDING),
        sg.Input(date.today().strftime('%Y-%m'), size=(8, 1), key="periodo", change_submits=True),
        sg.Text("De La:", pad=TEXT_PADDING),
        sg.Input("1", key="inicio", size=(4, 1), change_submits=True),
        sg.Text("Hasta:", pad=TEXT_PADDING),
        sg.Input("", key="final", size=(4, 1), change_submits=True),
        sg.Text("             ", pad=TEXT_PADDING),
        sg.Button("Exportar Excel", key="exportar_facturas", border_width=0),
        sg.Button("Ver Excel", key="ver_excel", border_width=0),
    ]

    c_second = [
        sg.Column(
            [
                [
                    sg.Button("Buscar Factura:", key="buscar_factura", border_width=0, pad=TEXT_PADDING),
                    sg.Input("", size=(40, 1), key="factura_pagar", change_submits=True),
                    sg.Button("Descarga", key="descarga", border_width=0, disabled=True),
                ],
                [
                    sg.Button("Status SAT", key="status_sat", border_width=0, disabled=True),
                    sg.Button("PUE Pagada", key="pago_pue", border_width=0, disabled=True),
                    sg.Button("PPD Ignorar", key="ignorar_ppd", border_width=0, disabled=True),
                    sg.Button("Email Notificada", key="email_notificada", border_width=0, disabled=True),
                    sg.Button(" Ver PDF ", key="ver_factura", border_width=0, disabled=True),
                ]
            ],
            pad=0
        ),
        sg.VSeparator(),
        sg.Column(
            [
                [
                    sg.CalendarButton("FechaPago:", format='%Y-%m-%d', title="FechaPago", no_titlebar=False, target="fecha_pago", pad=TEXT_PADDING, border_width=0,
                                      key="fecha_pago_select", disabled=True),
                    sg.Input("", size=(12, 1), key="fecha_pago", change_submits=True, disabled=True),
                    sg.Text("FormaPago:", pad=TEXT_PADDING),
                    sg.Combo([Code(k, v) for k, v in FORMA_PAGO.items()], default_value=Code("03", FORMA_PAGO["03"]), key="forma_pago", change_submits=True, size=(34, 1), disabled=True),
                ],
                [
                    sg.Button("Preparar Comprobante Pago", key="prepare_pago", border_width=0, disabled=True),
                ]
            ],
            pad=0
        )
    ]

    button_column_third = [
        sg.Button("Preparar Correos", key="prepare_correos", border_width=0),
        sg.Button("Facturas Pendientes", key="facturas_pendientes", border_width=0),
        sg.Checkbox("Ver Detallado", default=False, key="detallado"),
        sg.VSeparator(),
        sg.Button("Ajuste Anual", key="preparar_ajuste_anual", border_width=0),
        sg.Text("Año-Mes:", pad=TEXT_PADDING),
        sg.Input((date.today() + relativedelta(months=1)).strftime('%Y-%m'), size=(8, 1), key="anio_mes_ajuste"),
        sg.Text("Ajuste:", pad=TEXT_PADDING),
        sg.Input("", size=(6, 1), key="ajuste_porcentaje"),
        sg.Text("%", pad=RTEXT_PADDING),
        sg.VSeparator(),
        sg.Text("Recuperar:", pad=TEXT_PADDING),
        sg.Button("Emitidas", key="recuperar_emitidas", border_width=0, disabled=not has_fiel),
        sg.Button("Recibidas", key="recuperar_recibidas", border_width=0, disabled=not has_fiel),
        sg.Text("Dias:", pad=TEXT_PADDING),
        sg.Input("40", size=(4, 1), key="recuperar_dias", disabled=not has_fiel),
    ]

    button_column_low = [
        sg.Button("Validar Clientes", key="validate_clientes", border_width=0),
        sg.Button("Crear Facturas", disabled=True, key="crear_facturas", border_width=0),
        sg.Button("Enviar Correos", disabled=True, key="enviar_correos", border_width=0),
        sg.Push(),
        sg.Button("Acerca De", key="about", border_width=0),
    ]

    # ----- Full layout -----
    return [
        button_column,
        [sg.HSeparator()],
        c_second,
        [sg.HSeparator()],
        button_column_third,
        [sg.Output(expand_x=True, expand_y=True, key="console")],
        button_column_low
    ]


class InvoiceButtonManager:
    def __init__(self, button, detallado):
        self._cfdis = []
        self.button = button
        self.detallado = detallado

    def set_invoices(self, invoices):
        self._cfdis = invoices
        self.style_button()

    def clear(self):
        cfdis = self._cfdis
        self._cfdis = []
        self.style_button()
        return cfdis

    def style_button(self):
        self.button.update(
            f"Crear {len(self._cfdis)} Facturas" if self._cfdis else "Crear Facturas",
            disabled=len(self._cfdis) == 0
        )


class EmailButtonManager:
    def __init__(self, button):
        self._emails = {}
        self.button = button

    def set_emails(self, emails):
        self._emails = emails

        for i, (receptor, notify_invoices, facturas_pendientes) in enumerate(emails, start=1):
            log_item(f"CORREO NUMERO: {i}")
            log_email(receptor, notify_invoices, facturas_pendientes)

        self.style_button()

    def clear(self):
        emails = self._emails
        self._emails = {}
        self.style_button()
        return emails

    def style_button(self):
        self.button.update(
            f"Enviar {len(self._emails)} Correos" if self._emails else "Enviar Correos",
            disabled=len(self._emails) == 0
        )




import logging
from decimal import Decimal
from html import escape as html_escape

import jinja2
import yaml
from jinja2 import Environment
from jinja2.filters import do_mark_safe
from satcfdi import Code
from satcfdi.pacs import sat
from satcfdi.transform.helpers import Xint

# noinspection PyUnresolvedReferences
from satcfdi.transform.catalog import CATALOGS

REGIMEN_FISCAL = CATALOGS['{http://www.sat.gob.mx/sitio_internet/cfd/catalogos}c_RegimenFiscal']

logger = logging.getLogger()
sat_manager = sat.SAT()

environment_default = Environment(
    loader=jinja2.FileSystemLoader(searchpath=['templates']),
    autoescape=False,
    trim_blocks=True,
    lstrip_blocks=True,
    undefined=jinja2.StrictUndefined,
)


def finalize_html(val):
    return do_mark_safe(
        tag(html_escape(val), "b")
    )


environment_bold_escaped = Environment(
    loader=jinja2.FileSystemLoader(searchpath=['templates']),
    autoescape=True,
    trim_blocks=True,
    lstrip_blocks=True,
    undefined=jinja2.StrictUndefined,
    finalize=finalize_html
)


class LocalData(dict):
    file_source = None

    def __init__(self):
        with open(self.file_source, "r", encoding="utf-8") as fs:
            super().__init__(yaml.safe_load(fs))


class ConfigManager(LocalData):
    file_source = "config.yaml"

    def folio(self):
        return self["folio"]

    def serie(self):
        return self["serie"]

    def inc_folio(self):
        self["folio"] += 1
        self.save()

    def save(self):
        with open(self.file_source, "w", encoding="utf-8") as fs:
            yaml.dump_all([self], fs, Dumper=yaml.SafeDumper, encoding="utf-8", allow_unicode=True, sort_keys=False)


class ClientsManager(LocalData):
    file_source = "clients.yaml"

    def __init__(self):
        super().__init__()
        for k in self:
            self[k]["Rfc"] = k
            self[k]["RegimenFiscal"] = Code(self[k]["RegimenFiscal"], REGIMEN_FISCAL.get(self[k]["RegimenFiscal"]))


class FacturasManager(LocalData):
    file_source = "facturas.yaml"


def tag(text, tag):
    return '<' + tag + '>' + text + '</' + tag + '>'


yaml.SafeDumper.add_multi_representer(dict, lambda dumper, data: dumper.represent_dict(data))
yaml.SafeLoader.add_constructor("!decimal", lambda loader, node: Decimal(loader.construct_scalar(node)))


def represent_decimal(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', str(data))


def represent_str(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', str(data))


yaml.SafeDumper.add_representer(Decimal, represent_decimal)
yaml.SafeDumper.add_representer(Code, represent_str)
yaml.SafeDumper.add_representer(Xint, represent_str)



import time

from typing import Optional, Any
from azure.core.credentials import TokenCredential, AccessToken
from .microsoft_oauth2 import get_token, token_refresh

AZURE_CLI = "04b07795-8ddb-461a-bbee-02f9e1bf7b46"


class CacheInteractiveBrowserCredential(TokenCredential):
    def __init__(self, cache, login_hint=None, browser_name=None):
        self.cache = cache
        self.login_hint = login_hint
        self.browser_name = browser_name

    """Authenticates by requesting a token from the Azure CLI.

    This requires previously logging in to Azure via "az login", and will use the CLI's currently logged in identity.
    """

    def get_token(self, *scopes: str, claims: Optional[str] = None, tenant_id: Optional[str] = None, **kwargs: Any) -> AccessToken:
        if tenant_id is None:
            raise ValueError("Tenant is required")
        if len(scopes) != 1:
            raise ValueError("This credential requires exactly one scope per token request.")
        resource = scopes[0]

        token = self.cache.get(tenant_id)

        if not token:
            token = get_token(
                issuer_url="https://login.microsoftonline.com/organizations/",  # is 'https://sts.windows.net/f8cdef31-a31e-4b4a-93e4-5f571e91255a/'
                app_client_id=AZURE_CLI,
                scopes=['offline_access', 'openid', 'profile'],
                login_hint=self.login_hint,
                domain_hint=self.login_hint.split("@")[-1] if "@" in (self.login_hint or "") else None,
                browser_name=self.browser_name,
            )
            self.cache.set(tenant_id, token)  # expire=int(token["expires_in"])

        service_token = self.cache.get((tenant_id, resource))

        if not service_token:
            # Tenant properties
            service_token = token_refresh(
                token=token,
                issuer_url=f"https://sts.windows.net/{tenant_id}/",
                scopes=[resource, 'offline_access', 'openid', 'profile'],
            )

            self.cache.set((tenant_id, 'main'), service_token, expire=int(service_token["expires_in"]))

        # return service_token
        now = int(time.time())
        return AccessToken(service_token["access_token"], now + int(service_token["expires_in"]))



import os
import setuptools

about = {}
current_dir = os.path.abspath(os.path.dirname(__file__))
package_dir = 'satdigitalinvoice'

with open(os.path.join(current_dir, package_dir, "__version__.py"), "r", encoding="utf-8") as f:
    exec(f.read(), about)

package = about["__package__"]

packages = setuptools.find_packages(
    exclude=["contrib", "docs", "tests"],
)

version = os.environ.get('RELEASE_VERSION', about["__version__"])

setuptools.setup(
    name=package,  # installation
    version=version,
    author=about["__author__"],
    author_email=about["__author__"],
    description=about["__description__"],
    long_description=open('readme.rst', 'r', encoding='utf-8').read(),
    long_description_content_type="text/x-rst",
    url=about["__url__"],
    project_urls={
        "Documentation": about["__docs_url__"],
        "Source": about["__url__"],
        # "Changelog": about["__change_log_url__"],
    },
    license=about["__license__"],
    keywords=[
        'cfdi',
        'sat',
        'facturación',
        'comprobante',
        'retenciones',
        'nómina',
        'pagos',
        'carta porte',
        'contabilidad',
        'e-invoicing',
        'DIOT'
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: Spanish",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11"
    ],
    python_requires='>=3.10',
    package_dir={package: package_dir},
    packages=packages,
    package_data={
        package: [
            "markdown_styles/*",
        ],
    },
    install_requires=[
        'satcfdi==1.1.8',
        'diskcache',
        'num2words',
        'PyYAML',
        'babel',
        'azure-keyvault',
        'markdown2',
        'python-dateutil',
        'PySimpleGUI',
        'XlsxWriter'
    ],
    extras_require={
        'test': [
            'coverage',
            'pytest',
        ]
    }
)



import base64
import json

from azure.keyvault.secrets import SecretClient


class CacheSecretClient(SecretClient):
    def __init__(self, cache, encryption_key, vault_url, credential, **kwargs):
        self.cache = cache
        self.encryption_key = encryption_key
        super().__init__(vault_url, credential, **kwargs)

    def cache_get_secret(self, name):
        if res := self.cache.get(name):
            return res

        res = self.get_secret(name)
        self.cache[name] = res
        return res

    def get_secret_value(self, name):
        secret = self.cache_get_secret(name)

        value = secret.value
        content_type = secret.properties.content_type

        match content_type:
            case "application/json":
                return json.loads(value)

            case "base64":
                return base64.b64decode(value)

            case "encrypted":
                return self.encryption_key.decrypt(value.encode())

            case "encrypted/json":
                res = self.encryption_key.decrypt(value.encode())
                return json.loads(res)

        return value
